#ifndef OBJECT_H
#define OBJECT_H
#include <stdlib.h>

#include "Global.h"
#include "ObjectType.h"
#include "Util.h"
#include "MonsterIlk.h"


struct shObject;
typedef shVector <shObject *> shObjectVector;

int selectObjectsByFlag (shObjectVector *dest, shObjectVector *src, int flag);
int selectObjects (shObjectVector *dest, shObjectVector *src, shObjId ilk);
int selectObjects (shObjectVector *dest, shObjectVector *src,
                   shObjectType type);
int selectObjectsByFunction (shObjectVector *dest, shObjectVector *src,
                             int (shObject::*idfunc) (), int neg = 0);
int unselectObjectsByFunction (shObjectVector *dest, shObjectVector *src,
                               int (shObject::*idfunc) ());
int selectObjectsByCallback (shObjectVector *dest, shObjectVector *src,
                             int (*objfunc) (shObject *), int neg = 0);
int unselectObjectsByCallback (shObjectVector *dest, shObjectVector *src,
                               int (*objfunc) (shObject *));


struct shObject
{
    /* NOTE: don't forget to update constructor AND split () method
       when adding new fields to this class!
       Oh yeah and saveObject() and loadObject() too!
    */
    shObjId mIlkId;
    shObjectIlk *myIlk () { return AllIlks[mIlkId]; }

    int mCount;               // for mergeable objects
    int mCharges;             // for chargeable objects (maxhp for wreck)
    int mHP;                  // hit points
    int mDamage;              // corrosion / fire damage
    int mEnhancement;         // operating system for computers
    char mLetter;             // letter in Hero's inventory display
    int mBugginess;           // optimized +1 / debugged 0 / buggy -1
    shTime mLastEnergyBill;   // last time energy was billed to the item
    enum Flags {
        kKnownBugginess =   0x1,
        kKnownFooproof =    0x2,
        kKnownEnhancement = 0x4,   // so called plus
        kKnownCharges =     0x8,   // # of charges
        kKnownAppearance =  0x10,  // disk labels, canister types
        kKnownDamage =      0x20,  // disk/computer infection
        kKnownCracked =     0x40,  // hacked software
        kKnownExact =       0x7f,  // sum of all above
        /* 0x80 - unused yet */
        kFooproof =         0x100,  // acidproof, fireproof, etc.
                                    // also computers protected by antivirus
        kCracked =          0x200,  // floppy disks and keycards
        /* 0x400 - unused yet */
        /* 0x800 - unused yet */
        kWorn =             0x1000,
        kWielded =          0x2000,
        kActive =           0x4000, // turned on, for example
        kUnpaid =           0x8000, // belongs to shopkeeper
        kToggled =          0x10000,
        kDesignated =       0x20000 // primary item for some action
    };
    int mFlags;

    enum Location {
        kNowhere,
        kFloor,
        kInventory
    } mLocation;
    int mX;
    int mY;
    shCreature *mOwner;       /* creature with obj in inventory (use kUnpaid
                                 for obj owned by shopkeeper) */

    const char *mUserName;    /* name given by user with name command */
    union {
        shMonId mWreckIlk;
        shObjectIlk::Site mImplantSite;
    };

    //constructor:
    shObject () {
        mIlkId = kObjNothing; mCount = 0; mEnhancement = 0; mCharges = 0;
        mHP = 0; mLetter = 0; mBugginess = 0; mFlags = 0; mOwner = NULL;
        mDamage = 0; mLastEnergyBill = MAXTIME; mUserName = NULL;
    }
    shObject (shObjId id);

    void saveState (int fd);
    void loadState (int fd);

    const char *getDescription ();
    const char *getShortDescription ();
    const char *getVagueDescription ();
    //char *getFarDescription ();
    int isA (shObjId ilk);
    inline int isA (shObjectType type) { return myIlk ()->mType == type; }
    int getIlkFlag (int id) { return myIlk ()->mFlags & id; }
    int getFlag (int id) { return mFlags & id; }
    void setFlag (int id) { mFlags |= id; }
    void resetFlag (int id) { mFlags &= ~id; }
    int getMass () { return mCount * myIlk ()->mWeight; }
    void name (const char *newname = NULL);
    void nameIlk ();
    inline void maybeName ()
    {
        if (!isIlkKnown () && !myIlk ()->mUserName) nameIlk ();
    }
    inline int isIdentified () {
        return (kKnownExact == (kKnownExact & mFlags)) && isIlkKnown ();
    }
    inline int isChargeable () { return getIlkFlag (kChargeable); }
    inline int isChargeKnown () { return getFlag (kKnownCharges); }
    inline void setChargeKnown ()
    {
        setFlag (kKnownCharges);
        if (isA (kRayGun) && !mCharges)  mIlkId = kObjEmptyRayGun;
        if (mIlkId == kObjEmptyRayGun)   setIlkKnown ();
    }
    inline int isBugginessKnown () { return getFlag (kKnownBugginess); }
    inline void setBugginessKnown ()
    {
        setFlag (kKnownBugginess);
        if (mIlkId == kObjTorc) setFlag (kKnownAppearance);
    }
    inline void resetBugginessKnown () { resetFlag (kKnownBugginess); }
    inline int isAppearanceKnown () { return getFlag (kKnownAppearance); }
    inline void setAppearanceKnown ()
    {
        setFlag (kKnownAppearance);
        if (mIlkId == kObjTorc) setFlag (kKnownBugginess);
    }
    inline void resetAppearanceKnown () { resetFlag (kKnownAppearance); }
    inline int isDamageKnown () { return getFlag (kKnownDamage); }
    inline void setDamageKnown () { setFlag (kKnownDamage); }
    inline void resetDamageKnown () { resetFlag (kKnownDamage); }
    inline int isFooproofable () {
        return vulnerability () != kNoEnergy && (isA (kArmor) || isA (kWeapon));
    }
    inline int isFooproofKnown () { return getFlag (kKnownFooproof); }
    inline void setFooproofKnown () { setFlag (kKnownFooproof); }
    inline void resetFooproofKnown () { resetFlag (kKnownFooproof); }
    inline int isCrackedKnown () { return getFlag (kKnownCracked); }
    inline void setCrackedKnown () { setFlag (kKnownCracked); }
    inline int isCrackable () {
        return isA (kFloppyDisk) || isA (kObjGenericKeycard);
    }
    inline int isInfectedKnown () { return isDamageKnown (); }
    inline void setInfectedKnown () { setDamageKnown (); }
    inline void resetInfectedKnown () { resetDamageKnown (); }
    inline int isUnique () { return getIlkFlag (kUnique); }
    inline int isUniqueKnown () { return getIlkFlag (kUnique) && getIlkFlag (kIdentified); }
    inline int isIlkKnown () { return getIlkFlag (kIdentified); }
    inline void setIlkKnown ()
    {
        setAppearanceKnown ();
        if (!isIlkKnown ()) {
            if (myIlk ()->mUserName) {
                free ((void *) myIlk ()->mUserName);
                myIlk ()->mUserName = NULL;
            }
        }
        myIlk ()->mFlags |= kIdentified;
    }
    inline int isEnhancementKnown () { return getFlag (kKnownEnhancement); }
    inline void setEnhancementKnown () { setFlag (kKnownEnhancement); }
    inline void resetEnhancementKnown () { resetFlag (kKnownEnhancement); }
    inline int isEnhanceable () { return getIlkFlag (kEnhanceable); }
    inline int isRadioactive () {
        return (isA (kObjXRayGoggles) && isBuggy ()) ||
               (isA (kObjMutagenCanister) && !isOptimized ()) ||
               getIlkFlag (kRadioactive);
    }
    inline int isKnownRadioactive () {
        /* It is radioactive and you know it. */
        return (isRadioactive () && isIlkKnown () &&
                (!isA (kObjXRayGoggles) || isBugginessKnown ())) ||
        /* It is likely to be radioactive and you suspect it. */
               (isA (kObjMutagenCanister) && isIlkKnown () &&
                (!isBugginessKnown () || !isOptimized ()));
    }

    inline void identify () { setFlag (kKnownExact); setIlkKnown (); }

    shEnergyType vulnerability ();
    inline int isFooproof () { return getFlag (kFooproof); }
    inline void setFooproof () { setFlag (kFooproof); }
    inline void resetFooproof () { resetFlag (kFooproof); }
    inline int isIndestructible () { return getIlkFlag (kIndestructible); }
    int isFlammable () { return isA (kFloppyDisk); }
    inline int isBugProof () { return getIlkFlag (kBugProof); }
    inline int isBuggy () { return -1 == mBugginess; }
    inline int isDebugged () { return 0 == mBugginess; }
    inline int isOptimized () { return 1 == mBugginess; }
    inline void setBuggy () { if (!isBugProof ()) mBugginess = -1; }
    inline void setOptimized () { if (!isBugProof ()) mBugginess = 1; }
    inline void setDebugged () { mBugginess = 0; }
    inline void setBugginess (int b) { if (!isBugProof ()) mBugginess = b; }
    inline int isInfected () {
        return (isA (kFloppyDisk) || isComputer ()) && mDamage;
    }
    inline void setInfected () { mDamage = 1; }
    inline void resetInfected () { mDamage = 0; }
    /* TODO: Those functions below ask to be replaced with just is (), set ()
       and reset () just like conditions are for monsters. */
    inline int isCracked () { return getFlag (kCracked); }
    inline void setCracked () { setFlag (kCracked); }
    inline void resetCracked () { resetFlag (kCracked); }
    inline int isActive () { return getFlag (kActive); }
    inline void setActive () { setFlag (kActive); mLastEnergyBill = Clock; }
    inline void resetActive () {
        resetFlag (kActive);
        mLastEnergyBill = MAXTIME;
    }
    inline int isWorn () { return getFlag (kWorn); }
    inline void setWorn () { setFlag (kWorn); }
    inline void resetWorn () { resetFlag (kWorn); }
    inline int isWielded () { return getFlag (kWielded); }
    inline void setWielded () { setFlag (kWielded); }
    inline void resetWielded () { resetFlag (kWielded); }
    inline int isUnpaid () { return getFlag (kUnpaid); }
    inline void setUnpaid () { setFlag (kUnpaid); }
    inline void resetUnpaid () { resetFlag (kUnpaid); }
    inline int isToggled () { return getFlag (kToggled); }
    inline void setToggled () { setFlag (kToggled); }
    inline void resetToggled () { resetFlag (kToggled); }
    inline int isDesignated () { return getFlag (kDesignated); }
    inline void setDesignated () { setFlag (kDesignated); }
    inline void resetDesignated () { resetFlag (kDesignated); }

    inline int isDamaged () {
        return mDamage && !isA (kFloppyDisk) && !isComputer ();
    }
    inline int isEquipped () {
        return getFlag (kWorn) || getFlag (kWielded) || getFlag (kDesignated);
    }

    inline int canBeWorn () {
        return isA (kArmor) || isA (kImplant) || isComputer () ||
              (isA (kObjPlasmaCaster) && isIlkKnown () && isAppearanceKnown ());
    }
    inline int isSealedArmor () {
        return isA (kArmor) && getIlkFlag (kSealed);
    }
    inline int isPoweredArmor () {
        return isA (kArmor) && getIlkFlag (kPowered);
    }
    inline int isMeleeWeapon () {
        return NULL != myIlk ()->mMeleeAttack;
    }
    inline int isThrownWeapon () {
        return (isA(kWeapon) || isIlkKnown ()) && myIlk ()->mMissileAttack;
    }
    inline int isAimedWeapon () {
        return myIlk ()->mGunAttack || myIlk ()->mZapAttack;
    }
    inline int isWeldedWeapon () {
        return isWielded () && isBuggy () &&
            (isMeleeWeapon () || isAimedWeapon ());
    }
    inline int isSelectiveFireWeapon () {
        return isA (kWeapon) && getIlkFlag (kSelectiveFire);
    }
    inline int isKnownWeapon () {
        return isA (kWeapon) || (isIlkKnown () &&
            (myIlk ()->mMeleeAttack || myIlk ()->mGunAttack ||
             myIlk ()->mZapAttack));
    }
    inline int isComputer () {
        return isA (kObjGenericComputer);
    }

    inline int isUseable () {
        return myIlk ()->mUseFunc &&
            ((isA (kTool) || isA (kCanister)) ||
             (!isA (kImplant) && isIlkKnown ()) ||
             (isA (kImplant) && isIlkKnown () &&
              (isWorn () || isA (kObjMechaDendrites))));
    }
    int isAmmo (shObject *weapon);

    void impact (shCreature *c, shDirection dir,
                 shCreature *thrower = NULL, shObject *stack = NULL);
    void impact (int x, int y, shDirection dir,
                 shCreature *thrower = NULL, shObject *stack = NULL);
    void impact (shFeature *c, shDirection dir,
                 shCreature *thrower = NULL, shObject *stack = NULL);

    int sufferDamage (shAttack *attack, int x, int y, int multiplier = 1,
                      int specialonly = 0);

    int runFloppyDisk (shObject *computer);
    const shAbilities *getAbilityModifiers ();
    inline int getConferredIntrinsics () {
        return myIlk ()->mCarriedIntrinsics |
            (isWorn () ? myIlk ()->mWornIntrinsics : 0) |
            (isWielded () ? myIlk ()->mWieldedIntrinsics : 0) |
            (isActive () ? myIlk ()->mActiveIntrinsics : 0);
    }
    void applyConferredResistances (shCreature *target);
    void applyConferredSkills (shCreature *target);

    void draw (shInterface *I);
    void showLore (int analyze = 0);
    int canMerge (shObject *obj = NULL);
    void merge (shObject *obj); /* obj will be deleted */
    shObject *split (int count);

    const char *these () {
        char *buf = GetBuf ();
        if (mCount > 1) {
            snprintf (buf, SHBUFLEN, "these %d %s",
                      mCount, getDescription ());
        } else {
            snprintf (buf, SHBUFLEN, "this %s", getDescription ());
        }
        return buf;
    }

    const char *the () {
        char *buf = GetBuf ();
        if (mCount > 1) {
            snprintf (buf, SHBUFLEN, "the %d %s",
                      mCount, getDescription ());
        } else {
            snprintf (buf, SHBUFLEN, "the %s", getDescription ());
        }
        return buf;
    }

    const char *theQuick () {
        char *buf = GetBuf ();
        if (mCount > 1) {
            snprintf (buf, SHBUFLEN, "the %d %s",
                      mCount, getShortDescription ());
        } else {
            snprintf (buf, SHBUFLEN, "the %s", getShortDescription ());
        }
        return buf;
    }

    const char *an () {
        char *buf = GetBuf ();
        const char *tmp = getDescription ();

        if (mCount > 1) {
            snprintf (buf, SHBUFLEN, "%d %s", mCount, tmp);
        } else {
            snprintf (buf, SHBUFLEN, "%s %s",
                      isUniqueKnown () ? "the" :
                      isvowel (tmp[0]) ? "an" : "a", tmp);
        }
        return buf;
    }

    const char *anQuick () {
        char *buf = GetBuf ();
        const char *tmp = getShortDescription ();

        if (mCount > 1) {
            snprintf (buf, SHBUFLEN, "%d %s", mCount, tmp);
        } else {
            snprintf (buf, SHBUFLEN, "%s %s",
                      isUniqueKnown () ? "the" :
                      isvowel (tmp[0]) ? "an" : "a", tmp);
        }
        return buf;
    }

    const char *anVague () {
        char *buf = GetBuf ();
        const char *tmp = getVagueDescription ();
        if (mCount > 1) {
            snprintf (buf, SHBUFLEN, "%d %s", mCount, tmp);
        } else {
            snprintf (buf, SHBUFLEN, "%s %s",
                      isUniqueKnown () ? "the" :
                      isvowel (tmp[0]) ? "an" : "a", tmp);
        }
        return buf;
    }

    const char *inv ();

    const char *your () {
        char *buf = GetBuf ();
        if (mCount > 1) {
            snprintf (buf, SHBUFLEN, "your %d %s",
                      mCount, getShortDescription ());
        } else {
            snprintf (buf, SHBUFLEN, "your %s",
                      getShortDescription ());
        }
        return buf;
    }

    const char *her (shCreature *owner);

    /* any object can be used as a weapon, so we supply these routines: */
    int getThreatRange (shCreature *target);
    int getCriticalMultiplier ();

    int getArmorBonus ();
    int getPsiModifier ();

    int isXenosHunterGear ()
    {
        return mIlkId == kObjAquamarinePH || mIlkId == kObjMeanGreenPH ||
               mIlkId == kObjAquamarinePA || mIlkId == kObjMeanGreenPA ||
               mIlkId == kObjAdamantineCloak;
    }
};


int compareObjects (shObject **obj1, shObject **obj2);


int quaffCanister (shObject *can);
int executeFloppyDisk (shObject *computer, shObject *disk);

// infected == 1 means tell looks only
void identifyObjects (int howmany, int infected);

shObject *findBestObject (shObjectVector *objs);
shObject *createWreck (shCreature *m);

void makePlural (char *buf, int len);
shObject *createObject (const char *desc, int flags);

shObject *generateObject (int level = 1);
shObject *pickFromCategory (shObjId category);
shObject *createWeapon ();
shObject *createArmor ();
shObject *createTool ();
shObject *createCanister ();
shObject *createFloppyDisk ();
shObject *createImplant ();
shObject *createRayGun ();

void initializeMoney ();
void initializeWeapons ();
void initializeArmor ();
void initializeTools ();
void initializeRayGuns ();
void initializeCanisters ();
void initializeFloppyDisks ();
void initializeImplants ();
void initializeEnergy ();
void initializeWreck ();
void initializeArtifacts ();
void initializeObjects ();
void initializeLore ();
void initializeParser ();

void purgeDeletedObjects ();

void itemSpoilers ();

extern shObject *TheNothingObject;

#endif
